#
# Copyright (C) 2007, 2008 Red Hat, Inc.
# Authors:
# Thomas Woerner <twoerner@redhat.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os, os.path
import tempfile
import shutil
import types

from fw_config import CHAIN_NAME, _
from fw_functions import *

##############################################################################

class ip4tablesConfig:
    def __init__(self, filename):
        self.filename = filename
        self.clear()

    def clear(self):
        self.p_config = { }

    def get(self, key):
        if key in self.p_config.keys():
            return self.p_config[key]
        return None

    def set(self, key, value):
        if key[-8:] == "_MODULES":
            self.p_config[key.strip()] = value
        else:
            self.p_config[key.strip()] = value.strip()

    def __str__(self):
        s = ""
        for (key,value) in self.p_config.items():
            if s:
                s += '\n'
            s += '%s = %s' % (key, value)
        return s

    # load self.filename
    def read(self):
        self.clear()
        file = open(self.filename, "r")
        for line in file.xreadlines():
            if not line:
                break
            line = line.strip()
            if len(line) < 1 or line[0] == '#':
                continue
            # get key/value pairs
            p = line.split("=")
            if len(p) != 2:
                continue
            key = p[0].strip()
            value = p[1].strip()
            # remove leading and trailing double quotes
            if len(value) > 0 and value[0] == '"' and value[-1] == '"':
                value = value[1:-1]
            if key[-8:] == "_MODULES":
                value = value.split()
            self.p_config[key] = value
        file.close()

    # save to self.filename if there are key/value changes
    def write(self):
        if len(self.p_config) < 1:
            # no changes: nothing to do
            return

        (temp_file, temp) = tempfile.mkstemp("iptables-config")
        os.chmod(temp, 0600)
        modified = False
        file = open(self.filename, "r")
        for line in file.xreadlines():
            if not line: break
            # remove newline at and on line
            if line[-1:] == "\n":
                line = line[:-1]
            if len(line) < 1:
                os.write(temp_file, "\n")
                continue

            if line[0] != "#" and len(line) > 1:
                p = line.split("=")
                if len(p) != 2:
                    os.write(temp_file, line+"\n")
                    continue
                key = p[0].strip()
                value = p[1].strip()
                # remove leading and trailing double quotes
                if len(value) > 0 and value[0] == '"' and value[-1] == '"':
                    value = value[1:-1]
                if key[-8:] == "_MODULES":
                    value = value.split()
                if (key in self.p_config.keys() and \
                        self.p_config[key] != value) or \
                        key not in self.p_config.keys():
                    self._write(temp_file, key, self.p_config[key])
                    modified = True
                    del self.p_config[key]
                else:
                    os.write(temp_file, line+"\n")
                    del self.p_config[key]
            else:
                os.write(temp_file, line+"\n")
        # write remaining key/value pairs
        if len(self.p_config) > 0:
            os.write(temp_file, "\n")
        for (key,value) in self.p_config.items():
            self._write(temp_file, key, value)
            modified = True

        file.close()
        os.close(temp_file)

        if not modified: # not modified: remove tempfile
            os.remove(temp)
            return
        # copy tempfile
        try:
            shutil.copy(temp, self.filename)
        except:
            os.remove(temp)
            raise IOError, "Permission denied: '%s'" % self.filename
        else:
            os.remove(temp)
            os.chmod(self.filename, 0600)

    def _write(self, fd, key, value):
        if isinstance(value, types.ListType) or \
               isinstance(value, types.TupleType):
            val = " ".join(value)
        else:
            val = value
        os.write(fd, '%s="%s"\n' % (key, val))

##############################################################################

ip6tablesConfig = ip4tablesConfig

##############################################################################

class iptablesClass:
    prog = "iptables"

    def __init__(self, filename):
        self.filename = filename

    def write(self, conf):
        custom_mangle = [ ]
        custom_nat = [ ]
        custom_filter = [ ]
        if conf.custom_rules and len(conf.custom_rules) > 0:
            for (type, table, _filename) in conf.custom_rules:
                if type != "ipv4":
                    continue
                # ignore missing files
                if not os.path.exists(_filename) or \
                       not os.path.isfile(_filename):
                    continue
                if table == "mangle":
                    custom_mangle.append(_filename)
                elif table == "nat":
                    custom_nat.append(_filename)
                elif table == "filter":
                    custom_filter.append(_filename)

        fd = open(self.filename, "w")
        os.chmod(self.filename, 0600)

        fd.write("# Firewall configuration written by system-config-firewall\n")
        fd.write("# Manual customization of this file is not recommended.\n")

        # mangle table
        if len(custom_mangle) > 0:
            fd.write("*mangle\n")
            fd.write(":PREROUTING ACCEPT [0:0]\n")
            fd.write(":INPUT ACCEPT [0:0]\n")
            fd.write(":FORWARD ACCEPT [0:0]\n")
            fd.write(":OUTPUT ACCEPT [0:0]\n")
            fd.write(":POSTROUTING ACCEPT [0:0]\n")
            # add custom mangle rules
            for _filename in custom_mangle:
                catFile(fd, _filename)
            fd.write("COMMIT\n")

        # nat table
        if (conf.masq and len(conf.masq) > 0) or len(custom_nat) > 0:
            fd.write("*nat\n")
            fd.write(":PREROUTING ACCEPT [0:0]\n")
            fd.write(":OUTPUT ACCEPT [0:0]\n")
            fd.write(":POSTROUTING ACCEPT [0:0]\n")
            # masquerade
            if conf.masq:
                for dev in conf.masq:
                    fd.write("-A POSTROUTING -o %s -j MASQUERADE\n" % dev)
            # add custom nat rules
            if len(custom_nat) > 0:
                for _filename in custom_nat:
                    catFile(fd, _filename)
            fd.write("COMMIT\n")

        fd.write("*filter\n")
        fd.write(":INPUT ACCEPT [0:0]\n")
        fd.write(":FORWARD ACCEPT [0:0]\n")
        fd.write(":OUTPUT ACCEPT [0:0]\n")
        fd.write(":%s - [0:0]\n" % CHAIN_NAME)
        fd.write("-A INPUT -j %s\n" % CHAIN_NAME)

        # accept lo
        fd.write("-A %s -i lo -j ACCEPT\n" % CHAIN_NAME)

        if conf.trust:
            for dev in conf.trust:
                fd.write("-A %s -i %s -j ACCEPT\n" % (CHAIN_NAME, dev))

        # allow ICMP
        fd.write("-A %s -p icmp --icmp-type any -j ACCEPT\n" % CHAIN_NAME)

        # allow IPsec traffic (matches Lokkit 0.50 behavior)
        if not conf.no_ipsec:
            fd.write("-A %s -p 50 -j ACCEPT\n" % CHAIN_NAME)
            fd.write("-A %s -p 51 -j ACCEPT\n" % CHAIN_NAME)

        # allow mDNS
        if not conf.no_mdns:
            fd.write("-A %s -p udp --dport 5353 -d 224.0.0.251 -j ACCEPT\n" % \
                     CHAIN_NAME)

        # allow CUPS browsing
        if not conf.no_ipp:
            fd.write("-A %s -p udp -m udp --dport 631 -j ACCEPT\n" % CHAIN_NAME)
            fd.write("-A %s -p tcp -m tcp --dport 631 -j ACCEPT\n" % CHAIN_NAME)

        # note: RELATED is extremely important as it matches ICMP error messages
        fd.write("-A %s -m state --state ESTABLISHED,RELATED -j ACCEPT\n" % \
                 CHAIN_NAME)

        # open ports for new connections
        if conf.ports and len(conf.ports) > 0:
            for (ports, proto) in conf.ports:
                if len(ports) == 1:
                    dports = "%s" % ports[0]
                else:
                    dports = "%s:%s" % (ports[0], ports[1])
                fd.write("-A %s -m state --state NEW -m %s -p %s --dport %s "
                         "-j ACCEPT\n" % (CHAIN_NAME, proto, proto, dports))

        # FORWARD
        if (conf.trust and len(conf.trust) > 0) or \
                (conf.masq and len(conf.masq) > 0):
            # allow ICMP
            fd.write("-A FORWARD -p icmp --icmp-type any -j ACCEPT\n")
            # accept lo
            fd.write("-A FORWARD -i lo -j ACCEPT\n")
            # trusted interfaces
            if conf.trust:
                for dev in conf.trust:
                    fd.write("-A FORWARD -i %s -j ACCEPT\n" % dev)
            # allow to output to masqueraded interfaces (IPv4 only)
            if conf.masq:
                for dev in conf.masq:
                    fd.write("-A FORWARD -o %s -j ACCEPT\n" % dev)
            # accept established and related connections
            fd.write("-A FORWARD -m state --state ESTABLISHED,RELATED "
                     "-j ACCEPT\n")

        # add custom filter rules
        if len(custom_filter) > 0:
            for _filename in custom_filter:
                catFile(fd, _filename)

        # reject all other
        reject_type = "icmp-host-prohibited" # icmp-admin-prohibited
        fd.write("-A %s -j REJECT --reject-with %s\n" % (CHAIN_NAME,
                                                         reject_type))
        fd.write("-A FORWARD -j REJECT --reject-with %s\n" % reject_type)

        fd.write("COMMIT\n");
        fd.close()

    def _run(self, prog, arg, verbose=False):
        cmd = "%s %s %s" % (prog, self.prog, arg)
        if not verbose:
            cmd += " >/dev/null 2>&1"
        return os.system(cmd) >> 8

    def start(self, verbose=False):
        return self._run("/sbin/service", "start", verbose)

    def restart(self, verbose=False):
        return self._run("/sbin/service", "restart", verbose)

    def condrestart(self, verbose=False):
        return self._run("/sbin/service", "condrestart", verbose)

    def status(self, verbose=False):
        return self._run("/sbin/service", "status", verbose)

    def stop(self, verbose=False):
        return self._run("/sbin/service", "stop", verbose)

    def chkconfig_on(self, verbose=False):
        return self._run("/sbin/chkconfig", "on", verbose)

    def chkconfig_off(self, verbose=False):
        return self._run("/sbin/chkconfig", "off", verbose)

    def unlink(self):
        if os.path.exists(self.filename) and os.path.isfile(self.filename):
            os.unlink(self.filename)

##############################################################################

class ip6tablesClass(iptablesClass):
    prog = "ip6tables"

    def write(self, conf):
        custom_mangle = [ ]
        custom_nat = [ ]
        custom_filter = [ ]
        if conf.custom_rules and len(conf.custom_rules) > 0:
            for (type, table, _filename) in conf.custom_rules:
                if type != "ipv6":
                    continue
                if table == "mangle":
                    custom_mangle.append(_filename)
                elif table == "nat":
                    custom_nat.append(_filename)
                elif table == "filter":
                    custom_filter.append(_filename)

        fd = open(self.filename, "w")
        os.chmod(self.filename, 0600)

        fd.write("# Firewall configuration written by system-config-firewall\n")
        fd.write("# Manual customization of this file is not recommended.\n")

        # mangle table
        if len(custom_mangle) > 0:
            fd.write("*mangle\n")
            fd.write(":PREROUTING ACCEPT [0:0]\n")
            fd.write(":INPUT ACCEPT [0:0]\n")
            fd.write(":FORWARD ACCEPT [0:0]\n")
            fd.write(":OUTPUT ACCEPT [0:0]\n")
            fd.write(":POSTROUTING ACCEPT [0:0]\n")
            # add custom mangle rules
            for _filename in custom_mangle:
                catFile(fd, _filename)
            fd.write("COMMIT\n")

# no support for nat for netfilterv6 for now
#        # nat table
#        if (conf.masq and len(conf.masq) > 0) or len(custom_nat) > 0:
#            fd.write("*nat\n")
#            fd.write(":PREROUTING ACCEPT [0:0]\n")
#            fd.write(":OUTPUT ACCEPT [0:0]\n")
#            fd.write(":POSTROUTING ACCEPT [0:0]\n")
#            # masquerade
#            if conf.masq:
#                for dev in conf.masq:
#                    fd.write("-A POSTROUTING -o %s -j MASQUERADE\n" % dev)
#            # add custom nat rules
#            if len(custom_nat) > 0:
#                for _filename in custom_nat:
#                    catFile(fd, _filename)
#            fd.write("COMMIT\n")

        fd.write("*filter\n")
        fd.write(":INPUT ACCEPT [0:0]\n")
        fd.write(":FORWARD ACCEPT [0:0]\n")
        fd.write(":OUTPUT ACCEPT [0:0]\n")
        fd.write(":%s - [0:0]\n" % CHAIN_NAME)
        fd.write("-A INPUT -j %s\n" % CHAIN_NAME)

        # accept lo
        fd.write("-A %s -i lo -j ACCEPT\n" % CHAIN_NAME)

        if conf.trust:
            for dev in conf.trust:
                fd.write("-A %s -i %s -j ACCEPT\n" % (CHAIN_NAME, dev))

        # allow ICMP
        fd.write("-A %s -p ipv6-icmp -j ACCEPT\n" % CHAIN_NAME)

        # allow IPsec traffic (matches Lokkit 0.50 behavior)
        if not conf.no_ipsec:
            # use ipv6header match
            fd.write("-A %s -m ipv6header --header 50 -j ACCEPT\n" % CHAIN_NAME)
            fd.write("-A %s -m ipv6header --header 51 -j ACCEPT\n" % CHAIN_NAME)

        # allow mDNS
        if not conf.no_mdns:
            fd.write("-A %s -p udp --dport 5353 -d ff02::fb -j ACCEPT\n" % \
                     CHAIN_NAME)

        # allow CUPS browsing
        if not conf.no_ipp:
            fd.write("-A %s -p udp -m udp --dport 631 -j ACCEPT\n" % CHAIN_NAME)
            fd.write("-A %s -p tcp -m tcp --dport 631 -j ACCEPT\n" % CHAIN_NAME)

        # note: RELATED is extremely important as it matches ICMP error messages
        fd.write("-A %s -m state --state ESTABLISHED,RELATED -j ACCEPT\n" % \
                 CHAIN_NAME)

        # open ports for new connections
        if conf.ports and len(conf.ports) > 0:
            for (ports, proto) in conf.ports:
                if len(ports) == 1:
                    dports = "%s" % ports[0]
                else:
                    dports = "%s:%s" % (ports[0], ports[1])
                fd.write("-A %s -m state --state NEW -m %s -p %s --dport %s "
                         "-j ACCEPT\n" % (CHAIN_NAME, proto, proto, dports))

        # FORWARD
        if conf.trust and len(conf.trust) > 0:
# no support for nat for netfilterv6 for now
#        if (conf.trust and len(conf.trust) > 0) or \
#                (conf.masq and len(conf.masq) > 0):
            # allow ICMP
            fd.write("-A FORWARD -p ipv6-icmp -j ACCEPT\n")
            # accept lo
            fd.write("-A FORWARD -i lo -j ACCEPT\n")
            # trusted interfaces
            if conf.trust:
                for dev in conf.trust:
                    fd.write("-A FORWARD -i %s -j ACCEPT\n" % dev)
# no support for nat for netfilterv6 for now
#            # allow to output to masqueraded interfaces (IPv4 only)
#            if conf.masq:
#                for dev in conf.masq:
#                    fd.write("-A FORWARD -o %s -j ACCEPT\n" % dev)
            # accept established and related connections
            fd.write("-A FORWARD -m state --state ESTABLISHED,RELATED "
                     "-j ACCEPT\n")

        # add custom filter rules
        if len(custom_filter) > 0:
            for _filename in custom_filter:
                catFile(fd, _filename)

        # reject all other
        reject_type = "icmp6-adm-prohibited" # icmp6-port-unreachable
        fd.write("-A %s -j REJECT --reject-with %s\n" % (CHAIN_NAME,
                                                         reject_type))
        fd.write("-A FORWARD -j REJECT --reject-with %s\n" % reject_type)

        fd.write("COMMIT\n");
        fd.close()
