#
# Copyright (C) 2007, 2008 Red Hat, Inc.
# Authors:
# Thomas Woerner <twoerner@redhat.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from copy import copy
from optparse import Option, OptionValueError, OptionParser, Values, \
    BadOptionError
import socket
from fw_config import FIREWALL_TYPES, FIREWALL_TABLES
from fw_functions import getPortID, getPortRange, getServiceName
import os.path
import sys

def _check_selinuxmode(option, opt, value):
    if value not in [ "enforcing", "permissive", "disabled" ]:
        dict = { "option": opt, "value": value }
        raise OptionValueError(_("option %(option)s: invalid selinux "
                                 "mode '%(value)s'.") % dict)
    return value

def _check_port(option, opt, value):
    failure = False
    try:
        (ports, protocol) = value.split(":")
    except:
        failure = True
    dict = { "option": opt, "value": value, "protocol": protocol }
    if not failure:
        range = getPortRange(ports.strip())
        if range == -1:
            failure = True
        elif range == None:
            raise OptionValueError(_("option %(option)s: port range "
                                     "'%(value)s' is not unique.") % dict)
        elif len(range) == 2 and range[0] >= range[1]:
            raise OptionValueError(_("option %(option)s: '%(value)s' is not "
                                     "a valid range (start port >= end "
                                     "port).") % dict)
    if not failure:
        protocol = protocol.strip()
        if protocol not in [ "tcp", "udp" ]:
            raise OptionValueError(_("option %(option)s: '%(protocol)s' is "
                                     "not a valid protocol.") % dict)
    if failure:
        raise OptionValueError(_("option %(option)s: invalid port "
                                 "definition '%(value)s'.") % dict)
    return (range, protocol)

def _check_rulesfile(option, opt, value):
    type = "ipv4"
    table = "filter"
    splits = value.split(":", 1)
    if len(splits) > 1 and splits[0] in FIREWALL_TYPES:
        type = splits[0]
        splits = splits[1].split(":", 1)
    if len(splits) > 1 and splits[0] in FIREWALL_TABLES:
        table = splits[0]
        splits = splits[1].split(":", 1)
    filename = ":".join(splits)

    if type == "ipv6" and table == "nat":
        dict = { "option": opt, "value": value }
        raise OptionValueError(_("option %(option)s=%(value)s: ipv6 has no "
                                 "nat support.") % dict)
    
    return (type, table, filename)

def _append_unique(option, opt, value, parser, *args, **kwargs):
    vals = getattr(parser.values, option.dest)
    if vals and value in vals:
        return
    parser.values.ensure_value(option.dest, []).append(value)

class _Option(Option):
    TYPES = Option.TYPES + ("port", "rulesfile", "selinuxmode", )
    TYPE_CHECKER = copy(Option.TYPE_CHECKER)
    TYPE_CHECKER["port"] = _check_port
    TYPE_CHECKER["rulesfile"] = _check_rulesfile
    TYPE_CHECKER["selinuxmode"] = _check_selinuxmode

def _addStandardOptions(parser):
    parser.add_option("--enabled",
                      action="store_true", dest="enabled",
                      help=_("Enable firewall (default)"))
    parser.add_option("--disabled",
                      action="store_false", dest="enabled",
                      help=_("Disable firewall"))
    parser.add_option("--addmodule",
                      action="callback", dest="add_module", type="string",
                      metavar=_("<module>"),  callback=_append_unique,
                      help=_("Enable an iptables module"))
    parser.add_option("--removemodule",
                      action="callback", dest="remove_module", type="string",
                      metavar=_("<module>"), callback=_append_unique,
                      help=_("Disable an iptables module"))
    parser.add_option("-p", "--port",
                      action="callback", dest="ports", type="port",
                      metavar=_("<port>[-<port>]:<protocol>"),
                      callback=_append_unique,
                      help=_("Open specific ports in the firewall "
                             "(e.g, ssh:tcp)"))
    parser.add_option("--no-ipsec",
                      action="store_true", dest="no_ipsec",
                      help=_("Disable Internet Protocol Security (IPsec)"))
    parser.add_option("--no-ipp",
                      action="store_true", dest="no_ipp",
                      help=_("Disable Internet Printing Protocol (IPP)"))
    parser.add_option("--no-mdns",
                      action="store_true", dest="no_mdns",
                      help=_("Disable Multicast DNS (mDNS)"))
    parser.add_option("-t", "--trust",
                      action="callback", dest="trust", type="string",
                      metavar=_("<interface>"), callback=_append_unique,
                      help=_("Allow all traffic on the specified device"))
    parser.add_option("-m", "--masq",
                      action="callback", dest="masq", type="string",
                      metavar=_("<interface>"), callback=_append_unique,
                      help=_("Masquerades traffic from the specified device"))
    parser.add_option( "--high", "--medium",
                      action="store_true", dest="enabled",
                      help=_("Backwards compatibilty, aliased to --enabled"))
    parser.add_option("--custom-rules",
                      action="callback", dest="custom_rules", type="rulesfile",
                      metavar=_("[<type>:][<table>:]<filename>"),
                      callback=_append_unique,
                      help=_("Specify a custom rules file for inclusion in "
                             "the firewall, after the "
                             "default rules. Default type: ipv4, "
                             "default table: filter. "
                             "(Example: ipv4:filter:/etc/sysconfig/"
                             "ipv4_filter_addon)"))

def _addSELinuxOptions(parser):
    parser.add_option("--selinux",
                      action="store", dest="selinux", type="selinuxmode",
                      metavar=_("<mode>"),
                      help=_("Configure SELinux mode: enforcing, permissive "
                             "or disabled"))
    parser.add_option("--selinuxtype",
                      action="store", dest="selinuxtype", type="string",
                      metavar=_("<type>"),
                      help=_("Configure SELinux type: Usually targeted or "
                             "strict Policy"))

def _parse_args(parser, args, options=None):
    (_options, _args) = parser.parse_args(args, options)
    if len(_args) != 0:
        parser.error(_("no such option: %s") % " ".join(_args))
    return _options

class _OptionParser(OptionParser):
    # overload print_help: rhpl._ returns UTF-8
    def print_help(self, file=None):
        if file is None:
            file = sys.stdout

        str = self.format_help()
        if isinstance(str, unicode):
            encoding = self._get_encoding(file)
            str = str.encode(encoding, "replace")
        file.write(str)
    def print_usage(self, file=None):
        pass
    def _match_long_opt(self, opt):
        if self._long_opt.has_key(opt):
            return opt
        raise BadOptionError(opt)

def _gen_parser():
    parser = _OptionParser(add_help_option=False, option_class=_Option)
    return parser

def parseSysconfigArgs(args, options=None):
    parser = _gen_parser()
    _addStandardOptions(parser)
    return _parse_args(parser, args, options)

def parseSELinuxArgs(args, options=None):
    parser = _gen_parser()
    _addSELinuxOptions(parser)
    return _parse_args(parser, args, options)

def parseArgs(args=None, options=None):
    parser = _gen_parser()
    parser.add_option("-?", "--help", "--usage", action="help",
                      help=_("Show this help message"))
    parser.add_option("-q", "--quiet",
                      action="store_true", dest="quiet",
                      help=_("Run noninteractively; process only command-line "
                             "arguments"))
    parser.add_option("-v", "--verbose",
                      action="store_true", dest="verbose",
                      help=_("Be more verbose"))
    parser.add_option("-n", "--nostart",
                      action="store_true", dest="nostart",
                      help=_("Configure firewall but do not activate the new "
                             "configuration"))
    parser.add_option("-f",
                      action="store_true", dest="force",
                      help=_("Ignore actual settings"))
    parser.add_option("--update",
                      action="store_true", dest="update",
                      help=_("Update firewall non-interactively if the "
                             "firewall is enabled. This will also restart the "
                             "firewall. The -n and -f options will be "
                             "ignored."))
    _addSELinuxOptions(parser)
    _addStandardOptions(parser)

    return _parse_args(parser, args, options)

def copyValues(values):
    new_values = Values()
    new_values.__dict__ = copy(values.__dict__)
    return new_values
